<?php

namespace swf;

/*
    SWF.php: Macromedia Flash (SWF) file parser
    Copyright (C) 2012 Thanos Efraimidis (4real.gr)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/




class SWFtag {
    private SWFio $io; // SWFio for basic I/O
    private SWFrec $rec; // SWFrec for simple and complex records
    private int $swfVersion; // Version of this SWF file

    public function __construct(SWFio $io, SWFrec $rec, int $swfVersion) {
        $this->io = $io;
        $this->rec = $rec;
        $this->swfVersion = $swfVersion;
    }

    public function parseTags() {
        $tags = array();
        while ($this->io->bytePos < strlen($this->io->b)) {
            // Collect record header (short or long)
            $recordHeader = $this->io->collectUI16();
            $tagType = $recordHeader >> 6;
            $tagLength = $recordHeader & 0x3f;
            if ($tagLength == 0x3f) {
                $tagLength = $this->io->collectSI32();
            }
            // For definition tags, collect the 'id' also
            if ($this->isDefinitionTagType($tagType)) {
                $tags[] = array
                ('type' => $tagType,
                    'offset' => $this->io->bytePos,
                    'length' => $tagLength,
                    'id' => $this->io->collectUI16());
                $this->io->bytePos += $tagLength - 2; // 2 bytes already consumed
            } else {
                $tags[] = array
                ('type' => $tagType,
                    'offset' => $this->io->bytePos,
                    'length' => $tagLength);
                $this->io->bytePos += $tagLength;
            }
        }
        return $tags;
    }

    private function isDefinitionTagType($tagType) {
        switch ($tagType) {
            case  2: // DefineShape
            case 22: // DefineShape2
            case 32: // DefineShape3
            case 83: // DefineShape4
                return true; // shapeId
            case 10: // DefineFont
            case 48: // DefineFont2
            case 75: // DefineFont3
            case 91: // DefineFont4
                return true; // fontId
            case  7: // DefineButton
            case 34: // DefineButton2
                return true; // buttonId
            case 14: // DefineSound
                return true; // soundId
            case 39: // DefineSprite
                return true; // spriteId
            case 11: // DefineText
            case 33: // DefineText2
            case 20: // DefineBitsLossless
            case 36: // DefineBitsLossless2
            case  6: // DefineBits
            case 21: // DefineBitsJPEG2
            case 35: // DefineBitsJPEG3
            case 90: // DefineBitsJPEG4
            case 37: // DefineEditText
            case 46: // DefineMorphShape
            case 84: // DefineMorphShape2
            case 60: // DefineVideoStream
                return true; // characterId
        }
        return false;
    }

    public function parseTag($tag) {
        $tagType = $tag['type'];
        $tagOffset = $tag['offset'];
        $tagLength = $tag['length'];

        $this->io->bytePos = $tagOffset;
        $this->bitPos = 0;

        $bytePosEnd = $tagOffset + $tagLength;

        switch ($tagType) {
            case 0: // End
                $ret = $this->parseEndTag($bytePosEnd);
                break;
            case 1: // ShowFrame
                $ret = $this->parseShowFrameTag($bytePosEnd);
                break;
            case 2: // DefineShape
                $ret = $this->parseDefineShapeTag($bytePosEnd, 1);
                break;
            case 4: // PlaceObject
                $ret = $this->parsePlaceObjectTag($bytePosEnd);
                break;
            case 5: // RemoveObject
                $ret = $this->parseRemoveObjectTag($bytePosEnd, 1);
                break;
            case 6: // DefineBits
                $ret = $this->parseDefineBitsTag($bytePosEnd);
                break;
            case 7: // DefineButton
                $ret = $this->parseDefineButtonTag($bytePosEnd, 1);
                break;
            case 8: // JPEGTables
                $ret = $this->parseJPEGTablesTag($bytePosEnd);
                break;
            case 9: // SetBackgroundColor
                $ret = $this->parseSetBackgroundColorTag($bytePosEnd);
                break;
            case 10: // DefineFont
                $ret = $this->parseDefineFontTag($bytePosEnd);
                break;
            case 11: // DefineText
                $ret = $this->parseDefineTextTag($bytePosEnd, 1);
                break;
            case 12: // DoAction
                $ret = $this->parseDoActionTag($bytePosEnd);
                break;
            case 13: // DefineFontInfo
                $ret = $this->parseDefineFontInfoTag($bytePosEnd, 1);
                break;
            case 14: // DefineSound
                $ret = $this->parseDefineSoundTag($bytePosEnd);
                break;
            case 15: // StartSound
                $ret = $this->parseStartSoundTag($bytePosEnd, 1);
                break;
            case 17: // DefineButtonSound
                $ret = $this->parseDefineButtonSoundTag($bytePosEnd);
                break;
            case 18: // SoundStreamHead
                $ret = $this->parseSoundStreamHeadTag($bytePosEnd, 1);
                break;
            case 19: // SoundStreamBlock
                $ret = $this->parseSoundStreamBlockTag($bytePosEnd);
                break;
            case 20: // DefineBitsLossless
                $ret = $this->parseDefineBitsLosslessTag($bytePosEnd, 1);
                break;
            case 21: // DefineBitsJPEG2
                $ret = $this->parseDefineBitsJPEGTag($bytePosEnd, 2);
                break;
            case 22: // DefineShape2
                $ret = $this->parseDefineShapeTag($bytePosEnd, 2);
                break;
            case 23: // DefineButtonCxform
                $ret = $this->parseDefineButtonCxformTag($bytePosEnd);
                break;
            case 24: // Protect
                $ret = $this->parseProtectTag($bytePosEnd);
                break;
            case 26: // PlaceObject2
                $ret = $this->parsePlaceObject2Tag($bytePosEnd, $this->swfVersion);
                break;
            case 28: // RemoveObject2
                $ret = $this->parseRemoveObjectTag($bytePosEnd, 2);
                break;
            case 32: // DefineShape3
                $ret = $this->parseDefineShapeTag($bytePosEnd, 3);
                break;
            case 33: // DefineText2
                $ret = $this->parseDefineTextTag($bytePosEnd, 2);
                break;
            case 34: // DefineButton2
                $ret = $this->parseDefineButtonTag($bytePosEnd, 2);
                break;
            case 35: // DefineBitsJPEG3
                $ret = $this->parseDefineBitsJPEGTag($bytePosEnd, 3);
                break;
            case 36: // DefineBitsLossless2
                $ret = $this->parseDefineBitsLosslessTag($bytePosEnd, 2);
                break;
            case 37: // DefineEditText
                $ret = $this->parseDefineEditTextTag($bytePosEnd);
                break;
            case 39: // DefineSprite
                $ret = $this->parseDefineSpriteTag($bytePosEnd);
                break;
            case 43: // FrameLabel
                $ret = $this->parseFrameLabelTag($bytePosEnd);
                break;
            case 45: // SoundStreamHead2
                $ret = $this->parseSoundStreamHeadTag($bytePosEnd, 2);
                break;
            case 46: // DefineMorphShape
                $ret = $this->parseDefineMorphShapeTag($bytePosEnd, 1);
                break;
            case 48: // DefineFont2
                $ret = $this->parseDefineFont23Tag($bytePosEnd, 2);
                break;
            case 56: // ExportAssets
                $ret = $this->parseExportAssetsTag($bytePosEnd);
                break;
            case 57: // ImportAssets
                $ret = $this->parseImportAssetsTag($bytePosEnd, 1);
                break;
            case 58: // EnableDebugger
                $ret = $this->parseEnableDebuggerTag($bytePosEnd, 1);
                break;
            case 59: // DoInitAction
                $ret = $this->parseDoInitActionTag($bytePosEnd);
                break;
            case 60: // DefineVideoStream
                $ret = $this->parseDefineVideoStreamTag($bytePosEnd);
                break;
            case 61: // VideoFrame
                $ret = $this->parseVideoFrameTag($bytePosEnd);
                break;
            case 62: // DefineFontInfo2
                $ret = $this->parseDefineFontInfoTag($bytePosEnd, 2);
                break;
            case 64: // EnableDebugger2
                $ret = $this->parseEnableDebuggerTag($bytePosEnd, 2);
                break;
            case 65: // ScriptLimits
                $ret = $this->parseScriptLimitsTag($bytePosEnd);
                break;
            case 66: // SetTabIndex
                $ret = $this->parseSetTabIndexTag($bytePosEnd);
                break;
            case 69: // FileAttributes
                $ret = $this->parseFileAttributesTag($bytePosEnd);
                break;
            case 70: // PlaceObject3
                $ret = $this->parsePlaceObject3Tag($bytePosEnd);
                break;
            case 71: // ImportAssets2
                $ret = $this->parseImportAssetsTag($bytePosEnd, 2);
                break;
            case 73: // DefineFontAlignZones
                $ret = $this->parseDefineFontAlignZonesTag($bytePosEnd);
                break;
            case 74: // CSMTextSettings
                $ret = $this->parseCSMTextSettingsTag($bytePosEnd);
                break;
            case 75: // DefineFont3
                $ret = $this->parseDefineFont23Tag($bytePosEnd, 3);
                break;
            case 76: // SymbolClass
                $ret = $this->parseSymbolClassTag($bytePosEnd);
                break;
            case 77: // Metadata
                $ret = $this->parseMetadataTag($bytePosEnd);
                break;
            case 78: // DefineScalingGrid
                $ret = $this->parseDefineScalingGridTag($bytePosEnd);
                break;
            case 82: // DoABC
                $ret = $this->parseDoAbcTag($bytePosEnd);
                break;
            case 83: // DefineShape4
                $ret = $this->parseDefineShapeTag($bytePosEnd, 4);
                break;
            case 84: // DefineMorphShape2
                $ret = $this->parseDefineMorphShapeTag($bytePosEnd, 2);
                break;
            case 86: // DefineSceneAndFrameLabelData
                $ret = $this->parseDefineSceneAndFrameLabelDataTag($bytePosEnd);
                break;
            case 87: // DefineBinaryData
                $ret = $this->parseDefineBinaryDataTag($bytePosEnd);
                break;
            case 88: // DefineFontName
                $ret = $this->parseDefineFontNameTag($bytePosEnd);
                break;
            case 89: // StartSound2
                $ret = $this->parseStartSoundTag($bytePosEnd, 2);
                break;
            case 90: // DefineBitsJPEG4
                $ret = $this->parseDefineBitsJPEGTag($bytePosEnd, 4);
                break;
            case 91: // DefineFont4
                $ret = $this->parseDefineFont4Tag($bytePosEnd);
                break;
            case 41:
            case 253:
            case 255: // Undocumented
                $ret = array();
                $ret['bytes'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
                break;
            default:
                throw new \Exception(sprintf('Internal error: tagType=%d', $tagType));
        }
        if ($this->io->bytePos != $tagOffset + $tagLength) {
            // Make sure we have consumed whole tag
            $sb = '';
            $sb .= sprintf("Internal error: tagType=%d, %d bytes left:",
                $tagType, $tagOffset + $tagLength - $this->io->bytePos);
            for ($o = $this->io->bytePos; $o < $tagOffset + $tagLength; $o++) {
                $sb .= sprintf(" %02x", ord($this->io->b[$o]));
            }
            error_log($sb);
        }
        return $ret;
    }

    private function parseEndTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'EndTag';
        return $ret;
    }

    private function parseShowFrameTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'ShowFrame';
        return $ret;
    }

    private function parseDefineShapeTag($bytePosEnd, $shapeVersion) {
        $ret = array();
        if ($shapeVersion == 1) {
            $ret['tagType'] = 'DefineShape';
        } else if ($shapeVersion == 2) {
            $ret['tagType'] = 'DefineShape2';
        } else if ($shapeVersion == 3) {
            $ret['tagType'] = 'DefineShape3';
        } else if ($shapeVersion == 4) {
            $ret['tagType'] = 'DefineShape4';
        }
        if ($shapeVersion == 1 || $shapeVersion == 2 || $shapeVersion == 3) {
            $ret['shapeId'] = $this->io->collectUI16();
            $ret['shapeBounds'] = $this->rec->collectRect();
            $ret['shapes'] = $this->rec->collectShapeWithStyle($shapeVersion);
        } else if ($shapeVersion == 4) {
            $ret['shapeId'] = $this->io->collectUI16();
            $ret['shapeBounds'] = $this->rec->collectRect();
            $ret['edgeBounds'] = $this->rec->collectRect();
            $this->io->collectUB(5); // Reserved, must be 0
            $ret['usesFillWindingRule'] = $this->io->collectUB(1);
            $ret['usesNonScalingStrokes'] = $this->io->collectUB(1);
            $ret['usesScalingStrokes'] = $this->io->collectUB(1);
            $ret['shapes'] = $this->rec->collectShapeWithStyle($shapeVersion);
        }
        return $ret;
    }

    private function parsePlaceObjectTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'PlaceObject';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['depth'] = $this->io->collectUI16();
        $ret['matrix'] = $this->rec->collectMatrix();
        if ($this->io->bytePos < $bytePosEnd) {
            $ret['colorTransform'] = $this->rec->collectColorTransform(false);
        }
        return $ret;
    }

    private function parseRemoveObjectTag($bytePosEnd, $version) {
        $ret = array();
        if ($version == 1) {
            $ret['tagType'] = 'RemoveObject';
            $ret['characterId'] = $this->io->collectUI16();
            $ret['depth'] = $this->io->collectUI16();
        } else if ($version == 2) {
            $ret['tagType'] = 'RemoveObject2';
            $ret['depth'] = $this->io->collectUI16();
        }
        return $ret;
    }

    private function parseDefineBitsTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineBits';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['imageData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseDefineButtonTag($bytePosEnd, $version) {
        $ret = array();
        if ($version == 1) {
            $ret['tagType'] = 'DefineButton';
            $ret['buttonId'] = $this->io->collectUI16();
            $ret['characters'] = $this->rec->collectButtonRecords($version);
            $ret['actions'] = $this->rec->collectActionRecords($bytePosEnd);
        } else if ($version == 2) {
            $ret['tagType'] = 'DefineButton2';
            $ret['buttonId'] = $this->io->collectUI16();
            $this->io->collectUB(7); // Reserved, must be 0
            $ret['trackAsMenu'] = $this->io->collectUB(1);
            $here = $this->io->bytePos;
            $ret['actionOffset'] = $this->io->collectUI16();
            $ret['characters'] = $this->rec->collectButtonRecords($version);
            if ($ret['actionOffset'] != 0) {
                $ret['actions'] = $this->rec->collectButtonCondActions($bytePosEnd);
            }
        }
        return $ret;
    }

    private function parseJPEGTablesTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'JPEGTables';
        $ret['JPEGData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseSetBackgroundColorTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'SetBackgroundColor';
        $ret['backgroundColor'] = $this->rec->collectRGB();
        return $ret;
    }

    private function parseDefineFontTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineFont';
        $ret['fontId'] = $this->io->collectUI16();
        // Collect and push back 1st element of OffsetTable (this is numGlyphs * 2)
        $numGlyphs = $this->io->collectUI16() / 2;
        $this->io->bytePos -= 2;
        $ret['offsetTable'] = array();
        for ($i = 0; $i < $numGlyphs; $i++) {
            $ret['offsetTable'][] = $this->io->collectUI16();
        }
        $ret['glyphShapeData'] = array();
        for ($i = 0; $i < $numGlyphs; $i++) {
            $ret['glyphShapeData'][] = $this->rec->collectShape(1);
        }
        return $ret;
    }

    private function parseDefineTextTag($bytePosEnd, $textVersion) {
        $ret = array();
        $ret['tagType'] = $textVersion == 1 ? 'DefineText' : 'DefineText2';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['textBounds'] = $this->rec->collectRect();
        $ret['textMatrix'] = $this->rec->collectMatrix();
        $ret['glyphBits'] = $this->io->collectUI8();
        $ret['advanceBits'] = $this->io->collectUI8();
        $ret['textRecords'] = $this->rec->collectTextRecords($ret['glyphBits'], $ret['advanceBits'], $textVersion);
        return $ret;
    }

    private function parseDoActionTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DoAction';
        $ret['actions'] = $this->rec->collectActionRecords($bytePosEnd);
        return $ret;
    }

    private function parseDefineFontInfoTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'DefineFontInfo' : 'DefineFontInfo2';
        $ret['fontId'] = $this->io->collectUI16();
        $fontNameLen = $this->io->collectUI8();
        $ret['fontName'] = $this->io->collectBytes($fontNameLen);

        $this->io->collectUB(2); // Reserved
        $ret['fontFlagsSmallText'] = $this->io->collectUB(1);
        $ret['fontFlagsShiftJIS'] = $this->io->collectUB(1);
        $ret['fontFlagsANSI'] = $this->io->collectUB(1);
        $ret['fontFlagsItalic'] = $this->io->collectUB(1);
        $ret['fontFlagsBold'] = $this->io->collectUB(1);
        $ret['fontFlagsWideCodes'] = $this->io->collectUB(1);

        if ($version == 1) {
            $ret['codeTable'] = array();
            while ($this->io->bytePos < $bytePosEnd) {
                $ret['codeTable'][] = $ret['fontFlagsWideCodes'] ? $this->io->collectUI16() : $this->io->collectUI8();
            }
        } else if ($version == 2) {
            $ret['languageCode'] = $this->io->collectUI8();

            $ret['codeTable'] = array();
            while ($this->io->bytePos < $bytePosEnd) {
                $ret['codeTable'][] = $this->io->collectUI16();
            }
        }
        return $ret;
    }

    private function parseDefineSoundTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineSound';
        $ret['soundId'] = $this->io->collectUI16();
        $ret['soundFormat'] = $this->io->collectUB(4);
        $ret['soundRate'] = $this->io->collectUB(2);
        $ret['soundSize'] = $this->io->collectUB(1);
        $ret['soundType'] = $this->io->collectUB(1);
        $ret['soundSampleCount'] = $this->io->collectUI32();
        $ret['soundData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseStartSoundTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'StartSound' : 'StartSound2';
        if ($version == 1) {
            $ret['soundId'] = $this->io->collectUI16();
        } else if ($version == 2) {
            $ret['soundClassName'] = $this->io->collectString();
        }
        $ret['soundInfo'] = $this->rec->collectSoundInfo();
        return $ret;
    }

    private function parseDefineButtonSoundTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineButtonSound';
        $ret['buttonId'] = $this->io->collectUI16();
        $ret['buttonSoundChar0'] = $this->io->collectUI16();
        if ($ret['buttonSoundChar0'] != 0) {
            $ret['buttonSoundInfo0'] = $this->rec->collectSoundInfo();
        }
        $ret['buttonSoundChar1'] = $this->io->collectUI16();
        if ($ret['buttonSoundChar1'] != 0) {
            $ret['buttonSoundInfo1'] = $this->rec->collectSoundInfo();
        }
        $ret['buttonSoundChar2'] = $this->io->collectUI16();
        if ($ret['buttonSoundChar2'] != 0) {
            $ret['buttonSoundInfo2'] = $this->rec->collectSoundInfo();
        }
        $ret['buttonSoundChar3'] = $this->io->collectUI16();
        if ($ret['buttonSoundChar3'] != 0) {
            $ret['buttonSoundInfo3'] = $this->rec->collectSoundInfo();
        }
        return $ret;
    }

    private function parseSoundStreamHeadTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'SoundStreamHead' : 'SoundStreamHead2';
        $this->io->collectUB(4); // Reserved
        $ret['playbackSoundRate'] = $this->io->collectUB(2);
        $ret['playbackSoundSize'] = $this->io->collectUB(1);
        $ret['playbackSoundType'] = $this->io->collectUB(1);

        $ret['streamSoundCompression'] = $this->io->collectUB(4);
        $ret['streamSoundRate'] = $this->io->collectUB(2);
        $ret['streamSoundSize'] = $this->io->collectUB(1);
        $ret['streamSoundType'] = $this->io->collectUB(1);

        $ret['streamSoundSampleCount'] = $this->io->collectUI16();
        if ($ret['streamSoundCompression'] == 2) { // MP3
            $ret['latencySeek'] = $this->io->collectSI16();
        }
        return $ret;
    }

    private function parseSoundStreamBlockTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'SoundStreamBlock';
        $ret['soundStreamData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseDefineBitsLosslessTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'DefineBitsLossless' : 'DefineBitsLossless2';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['bitmapFormat'] = $this->io->collectUI8();
        $ret['bitmapWidth'] = $this->io->collectUI16();
        $ret['bitmapHeight'] = $this->io->collectUI16();
        if ($ret['bitmapFormat'] == 3) {
            $colors = $this->io->collectUI8();
        }
        $data = gzuncompress($this->io->collectBytes($bytePosEnd - $this->io->bytePos)); // ZLIB uncompress
        if ($ret['bitmapFormat'] == 3) {
            // Colormap
            if ($version == 1) {
                $colorTableSize = 3 * ($colors + 1); // 3 bytes per RGB value
            } else if ($version == 2) {
                $colorTableSize = 4 * ($colors + 1); // 4 bytes per RGBA value
            }
            $ret['colorTable'] = substr($data, 0, $colorTableSize);
            $ret['pixelData'] = substr($data, $colorTableSize);
        } else if ($ret['bitmapFormat'] == 4 || $ret['bitmapFormat'] == 5) {
            $ret['pixelData'] = $data;
        } else {
            throw new \Exception(sprintf('Internal error: bitmapFormat=%d', $ret['bitmapFormat']));
        }
        return $ret;
    }

    private function parseDefineBitsJPEGTag($bytePosEnd, $version) {
        $ret = array();
        if ($version == 2) {
            $ret['tagType'] = 'DefineBitsJPEG2';
            $ret['characterId'] = $this->io->collectUI16();
            $ret['imageData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        } else if ($version == 3) {
            $ret['tagType'] = 'DefineBitsJPEG3';
            $ret['characterId'] = $this->io->collectUI16();
            $alphaDataOffset = $this->io->collectUI32();
            $ret['imageData'] = $this->io->collectBytes($alphaDataOffset);
            $ret['alphaData'] = gzuncompress($this->io->collectBytes($bytePosEnd - $this->io->bytePos)); // ZLIB uncompress alpha channel
        } else if ($version == 4) {
            $ret['tagType'] = 'DefineBitsJPEG4';
            $ret['characterId'] = $this->io->collectUI16();
            $alphaDataOffset = $this->io->collectUI32();
            $ret['deblockParam'] = $this->io->collectUI16();
            $ret['imageData'] = $this->io->collectBytes($alphaDataOffset);
            $ret['alphaData'] = gzuncompress($this->io->collectBytes($bytePosEnd - $this->io->bytePos)); // ZLIB uncompress alpha channel
        } else {
            throw new \Exception(sprintf('Internal error: version=%d', $version));
        }
        return $ret;
    }

    private function parseDefineButtonCxformTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineButtonCxform';
        $ret['buttonId'] = $this->io->collectUI16();
        $ret['colorTransform'] = $this->rec->collectColorTransform(false);
        return $ret;
    }

    private function parseProtectTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'Protect';
        //XXX a MD5 password may be present here
        return $ret;
    }

    private function parsePlaceObject2Tag($bytePosEnd, $swfVersion) {
        $ret = array();
        $ret['tagType'] = 'PlaceObject2';

        $ret['placeFlagHasClipActions'] = $this->io->collectUB(1);
        $ret['placeFlagHasClipDepth'] = $this->io->collectUB(1);
        $ret['placeFlagHasName'] = $this->io->collectUB(1);
        $ret['placeFlagHasRatio'] = $this->io->collectUB(1);
        $ret['placeFlagHasColorTransform'] = $this->io->collectUB(1);
        $ret['placeFlagHasMatrix'] = $this->io->collectUB(1);
        $ret['placeFlagHasCharacter'] = $this->io->collectUB(1);
        $ret['placeFlagMove'] = $this->io->collectUB(1);

        $ret['depth'] = $this->io->collectUI16();
        if ($ret['placeFlagHasCharacter'] != 0) {
            $ret['characterId'] = $this->io->collectUI16();
        }
        if ($ret['placeFlagHasMatrix'] != 0) {
            $ret['matrix'] = $this->rec->collectMatrix();
        }
        if ($ret['placeFlagHasColorTransform'] != 0) {
            $ret['colorTransform'] = $this->rec->collectColorTransform(true);
        }
        if ($ret['placeFlagHasRatio'] != 0) {
            $ret['ratio'] = $this->io->collectUI16();
        }
        if ($ret['placeFlagHasName'] != 0) {
            $ret['name'] = $this->io->collectString();
        }
        if ($ret['placeFlagHasClipDepth'] != 0) {
            $ret['clipDepth'] = $this->io->collectUI16();
        }
        if ($ret['placeFlagHasClipActions'] != 0) {
            $ret['clipActions'] = $this->rec->collectClipActions($swfVersion);
        }
        return $ret;
    }

    private function parseDefineEditTextTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineEditText';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['bounds'] = $this->rec->collectRect();

        $ret['hasText'] = $this->io->collectUB(1);
        $ret['wordWrap'] = $this->io->collectUB(1);
        $ret['multiline'] = $this->io->collectUB(1);
        $ret['password'] = $this->io->collectUB(1);
        $ret['readOnly'] = $this->io->collectUB(1);
        $ret['hasTextColor'] = $this->io->collectUB(1);
        $ret['hasMaxLength'] = $this->io->collectUB(1);
        $ret['hasFont'] = $this->io->collectUB(1);

        $ret['hasFontClass'] = $this->io->collectUB(1);
        $ret['autoSize'] = $this->io->collectUB(1);
        $ret['hasLayout'] = $this->io->collectUB(1);
        $ret['noSelect'] = $this->io->collectUB(1);
        $ret['border'] = $this->io->collectUB(1);
        $ret['wasStatic'] = $this->io->collectUB(1);
        $ret['HTML'] = $this->io->collectUB(1);
        $ret['useOutlines '] = $this->io->collectUB(1);

        if ($ret['hasFont'] != 0) {
            $ret['fontId'] = $this->io->collectUI16();
        }
        if ($ret['hasFontClass'] != 0) {
            $ret['fontClass'] = $this->io->collectString();
        }
        if ($ret['hasFont'] != 0) {
            $ret['fontHeight'] = $this->io->collectUI16();
        }
        if ($ret['hasTextColor'] != 0) {
            $ret['textColor'] = $this->rec->collectRGBA();
        }
        if ($ret['hasMaxLength'] != 0) {
            $ret['maxLength'] = $this->io->collectUI16();
        }
        if ($ret['hasLayout'] != 0) {
            $ret['align'] = $this->io->collectUI8();
            $ret['leftMargin'] = $this->io->collectUI16();
            $ret['rightMargin'] = $this->io->collectUI16();
            $ret['indent'] = $this->io->collectUI16();
            $ret['leading'] = $this->io->collectSI16();
        }
        $ret['variableName'] = $this->io->collectString();
        if ($ret['hasText'] != 0) {
            $ret['initialText'] = $this->io->collectString();
        }
        return $ret;
    }

    private function parseDefineSpriteTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineSprite';
        $ret['spriteId'] = $this->io->collectUI16();
        $ret['frameCount'] = $this->io->collectUI16();
        $b = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);

        $io = new SWFio($b);
        $rec = new SWFrec($io);
        $tag = new SWFtag($io, $rec, $this->swfVersion);

        // Collect and parse tags
        $ret['tags'] = array();
        while ($io->bytePos < strlen($io->b)) {
            $recordHeader = $io->collectUI16();
            $tagType = $recordHeader >> 6;
            $tagLength = $recordHeader & 0x3f;
            if ($tagLength == 0x3f) {
                $tagLength = $io->collectSI32();
            }
            $bytePosEnd = $io->bytePos + $tagLength;
            $ret['tags'][] = $tag->parseTag(array('type' => $tagType, 'offset' => $io->bytePos, 'length' => $tagLength));
            $io->bytePos = $bytePosEnd;
        }
        return $ret;
    }

    private function parseFrameLabelTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'FrameLabel';
        $ret['frameLabel'] = $this->io->collectString();
        return $ret;
    }

    private function parseDefineMorphShapeTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'DefineMorphShape' : 'DefineMorphShape2';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['startBounds'] = $this->rec->collectRect();
        $ret['endBounds'] = $this->rec->collectRect();
        if ($version == 2) {
            $ret['startEdgeBounds'] = $this->rec->collectRect();
            $ret['endEdgeBounds'] = $this->rec->collectRect();
            $this->io->collectUB(6); // Reserved
            $ret['[usesNonScalingStrokes'] = $this->io->collectUB(1);
            $ret['usesScalingStrokes'] = $this->io->collectUB(1);
        }
        $ret['offset'] = $this->io->collectUI32();
        $ret['morphFillStyles'] = $this->rec->collectMorphFillStyleArray();
        $ret['morphLineStyles'] = $this->rec->collectMorphLineStyleArray($version);
        $ret['startEdges'] = $this->rec->collectShape(1);
        $ret['endEdges'] = $this->rec->collectShape(1);
        return $ret;
    }

    private function parseDefineFont23Tag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 2 ? 'DefineFont2' : 'DefineFont3';
        $ret['fontId'] = $this->io->collectUI16();
        $ret['fontFlagsHasLayout'] = $this->io->collectUB(1);
        $ret['fontFlagsShiftJIS'] = $this->io->collectUB(1);
        $ret['fontFlagsSmallText'] = $this->io->collectUB(1);
        $ret['fontFlagsANSI'] = $this->io->collectUB(1);
        $ret['fontFlagsWideOffsets'] = $this->io->collectUB(1);
        $ret['fontFlagsWideCodes'] = $this->io->collectUB(1);
        $ret['fontFlagsItalic'] = $this->io->collectUB(1);
        $ret['fontFlagsBold'] = $this->io->collectUB(1);
        $ret['languageCode'] = $this->io->collectUI8();
        $fontNameLength = $this->io->collectUI8();
        $ret['fontName'] = substr($this->io->collectBytes($fontNameLength), 0, -1); // Remove trailing NULL
        $ret['numGlyphs'] = $this->io->collectUI16();

        $numGlyphs = $ret['numGlyphs'];
        $fontFlagsHasLayout = $ret['fontFlagsHasLayout'] != 0;
        $fontFlagsWideOffsets = $ret['fontFlagsWideOffsets'] != 0;
        $fontFlagsWideCodes = $ret['fontFlagsWideCodes'] != 0;
        $offsetTable = array();
        for ($i = 0; $i < $numGlyphs; $i++) {
            $offsetTable[] = $fontFlagsWideOffsets ? $this->io->collectUI32() : $this->io->collectUI16();
        }
        $codeTableOffset = $fontFlagsWideOffsets ? $this->io->collectUI32() : $this->io->collectUI16();
        $ret['glyphShapeTable'] = array();
        for ($i = 0; $i < $numGlyphs; $i++) {
            $ret['glyphShapeTable'][] = $this->rec->collectShape(1);
        }
        $ret['codeTable'] = array();
        for ($i = 0; $i < $numGlyphs; $i++) {
            if ($version == 2) {
                $ret['codeTable'][] = $fontFlagsWideCodes ? $this->io->collectUI16() : $this->io->collectUI8();
            } else if ($version == 3) {
                $ret['codeTable'][] = $this->io->collectUI16();
            }
        }

        if ($fontFlagsHasLayout) {
            $ret['fontAscent'] = $this->io->collectSI16();
            $ret['fontDescent'] = $this->io->collectSI16();
            $ret['fontLeading'] = $this->io->collectSI16();
            $ret['fontAdvanceTable'] = array();
            for ($i = 0; $i < $numGlyphs; $i++) {
                $ret['fontAdvanceTable'][] = $this->io->collectSI16();
            }
            $ret['fontBoundsTable'] = array();
            for ($i = 0; $i < $numGlyphs; $i++) {
                $ret['fontBoundsTable'][] = $this->rec->collectRect();
            }
            $kerningCount = $this->io->collectUI16();
            $ret['fontKerningTable'] = array();
            for ($i = 0; $i < $kerningCount; $i++) {
                $fontKerningCode1 = $fontFlagsWideCodes ? $this->io->collectUI16() : $this->io->collectUI8();
                $fontKerningCode2 = $fontFlagsWideCodes ? $this->io->collectUI16() : $this->io->collectUI8();
                $fontKerningAdjustment = $this->io->collectSI16();
                $ret['fontKerningTable'][] =
                    array('fontKerningCode1' => $fontKerningCode1,
                        'fontKerningCode2' => $fontKerningCode2,
                        'fontKerningAdjustment' => $fontKerningAdjustment);
            }
        }
        return $ret;
    }

    private function parseExportAssetsTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'ExportAssets';
        $ret['tags'] = array();
        $ret['names'] = array();
        $count = $this->io->collectUI16();
        for ($i = 0; $i < $count; $i++) {
            $ret['tags'][] = $this->io->collectUI16();
            $ret['names'][] = $this->io->collectString();
        }
        return $ret;
    }

    private function parseImportAssetsTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'ImportAssets' : 'ImportAssets2';
        $ret['URL'] = $this->io->collectString();
        if ($version == 2) {
            $this->io->collectUI8(); // Reserved, must be 1
            $this->io->collectUI8(); // Reserved, must be 0
        }
        $ret['tags'] = array();
        $ret['names'] = array();
        $count = $this->io->collectUI16();
        for ($i = 0; $i < $count; $i++) {
            $ret['tags'][] = $this->io->collectUI16();
            $ret['names'][] = $this->io->collectString();
        }
        return $ret;
    }

    private function parseEnableDebuggerTag($bytePosEnd, $version) {
        $ret = array();
        $ret['tagType'] = $version == 1 ? 'EnableDebugger' : 'EnableDebugger2';
        if ($version == 2) {
            $this->io->collectUI16(); // Reserved, must be 0
        }
        $ret['password'] = $this->io->collectString();
        return $ret;
    }

    private function parseDoInitActionTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DoInitAction';
        $ret['spriteId'] = $this->io->collectUI16();
        $ret['actions'] = $this->rec->collectActionRecords($bytePosEnd);
        return $ret;
    }

    private function parseDefineVideoStreamTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineVideoStream';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['numFrames'] = $this->io->collectUI16();
        $ret['width'] = $this->io->collectUI16();
        $ret['height'] = $this->io->collectUI16();

        $this->io->collectUB(4); // Reserved
        $ret['videoFlagsDeblocking'] = $this->io->collectUB(3);
        $ret['videoFlagsSmoothing'] = $this->io->collectUB(1);

        $ret['codecId'] = $this->io->collectUI8();
        return $ret;
    }

    private function parseVideoFrameTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'VideoFrame';
        $ret['streamId'] = $this->io->collectUI16();
        $ret['frameNum'] = $this->io->collectUI16();
        $ret['videoData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseScriptLimitsTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'ScriptLimits';
        $ret['maxRecursionDepth'] = $this->io->collectUI16();
        $ret['scriptTimeoutSeconds'] = $this->io->collectUI16();
        return $ret;
    }

    private function parseSetTabIndexTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'SetTabIndex';
        $ret['depth'] = $this->io->collectUI16();
        $ret['tabIndex'] = $this->io->collectUI16();
        return $ret;
    }

    private function parseFileAttributesTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'FileAttributes';
        $this->io->collectUB(1); // Reserved
        $ret['useDirectBlit'] = $this->io->collectUB(1);
        $ret['useGPU'] = $this->io->collectUB(1);
        $ret['hasMetadata'] = $this->io->collectUB(1);
        $ret['actionScript3'] = $this->io->collectUB(1);
        $this->io->collectUB(2); // Reserved
        $ret['useNetwork'] = $this->io->collectUB(1);
        $this->io->collectUB(24); // Reserved
        return $ret;
    }

    private function parsePlaceObject3Tag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'PlaceObject3';

        $ret['placeFlagHasClipActions'] = $this->io->collectUB(1);
        $ret['placeFlagHasClipDepth'] = $this->io->collectUB(1);
        $ret['placeFlagHasName'] = $this->io->collectUB(1);
        $ret['placeFlagHasRatio'] = $this->io->collectUB(1);
        $ret['placeFlagHasColorTransform'] = $this->io->collectUB(1);
        $ret['placeFlagHasMatrix'] = $this->io->collectUB(1);
        $ret['placeFlagHasCharacter'] = $this->io->collectUB(1);
        $ret['placeFlagMove'] = $this->io->collectUB(1);

        $this->io->collectUB(3); // Reserved, must be 0
        $ret['placeFlagHasImage'] = $this->io->collectUB(1);
        $ret['placeFlagHasClassName'] = $this->io->collectUB(1);
        $ret['placeFlagHasCacheAsBitmap'] = $this->io->collectUB(1);
        $ret['placeFlagHasBlendMode'] = $this->io->collectUB(1);
        $ret['placeFlagHasFilterList'] = $this->io->collectUB(1);

        $ret['depth'] = $this->io->collectUI16();
        if ($ret['placeFlagHasClassName'] != 0 || ($ret['placeFlagHasImage'] != 0 && $ret['placeFlagHasCharacter'] != 0)) {
            $ret['className'] = $this->io->collectString();
        }
        if ($ret['placeFlagHasCharacter'] != 0) {
            $ret['characterId'] = $this->io->collectUI16();
        }
        if ($ret['placeFlagHasMatrix'] != 0) {
            $ret['matrix'] = $this->rec->collectMatrix();
        }
        if ($ret['placeFlagHasColorTransform'] != 0) {
            $ret['colorTransform'] = $this->rec->collectColorTransform(true);
        }
        if ($ret['placeFlagHasRatio'] != 0) {
            $ret['ratio'] = $this->io->collectUI16();
        }
        if ($ret['placeFlagHasName'] != 0) {
            $ret['name'] = $this->io->collectString();
        }
        if ($ret['placeFlagHasClipDepth'] != 0) {
            $ret['clipDepth'] = $this->io->collectUI16();
        }
        if ($ret['placeFlagHasFilterList'] != 0) {
            $ret['surfaceFilterlist'] = $this->rec->collectFilterList();
        }
        if ($ret['placeFlagHasBlendMode'] != 0) {
            $ret['blendMode'] = $this->io->collectUI8();
        }
        if ($ret['placeFlagHasCacheAsBitmap'] != 0) {
            $ret['bitmapCache'] = $this->io->collectUI8();
        }
        if ($ret['placeFlagHasClipActions'] != 0) {
            $ret['clipActions'] = $this->rec->collectClipActions($this->swfVersion);
        }
        return $ret;
    }

    private function parseDefineFontAlignZonesTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineFontAlignZones';
        $ret['fontId'] = $this->io->collectUI16();
        $ret['CSMTableHint'] = $this->io->collectUB(2);
        $this->io->collectUB(6); // Reserved
        $ret['zoneTable'] = $this->rec->collectZoneTable($bytePosEnd);
        return $ret;
    }

    private function parseCSMTextSettingsTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'CSMTextSettings';
        $ret['textId'] = $this->io->collectUI16();
        $ret['useFlashType'] = $this->io->collectUB(2);
        $ret['gridFit'] = $this->io->collectUB(3);
        $this->io->collectUB(3); // Reserved
        $ret['thickness'] = $this->io->collectFixed(); //XXX F32 in the spec
        $ret['sharpness'] = $this->io->collectFixed(); //XXX F32 in the spec
        $this->io->collectUI8(); // Reserved
        return $ret;
    }

    private function parseSymbolClassTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'SymbolClass';
        $numSymbols = $this->io->collectUI16();
        $tags = array();
        $names = array();
        for ($i = 0; $i < $numSymbols; $i++) {
            $tags[] = $this->io->collectUI16();
            $names[] = $this->io->collectString();
        }
        $ret['tags'] = $tags;
        $ret['names'] = $names;
        return $ret;
    }

    private function parseMetadataTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'Metadata';
        $ret['metadata'] = $this->io->collectString();
        return $ret;
    }

    private function parseDefineScalingGridTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineScalingGrid';
        $ret['characterId'] = $this->io->collectUI16();
        $ret['splitter'] = $this->rec->collectRect();
        return $ret;
    }

    private function parseDoABCTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DoABC';
        $ret['flags'] = $this->io->collectUI32();
        $ret['name'] = $this->io->collectString();
        $ret['ABCdata'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseDefineSceneAndFrameLabelDataTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineSceneAndFrameLabelData';

        $sceneOffsets = array();
        $sceneNames = array();
        $sceneCount = $this->io->collectEncodedU32();
        for ($i = 0; $i < $sceneCount; $i++) {
            $sceneOffsets[] = $this->io->collectEncodedU32();
            $sceneNames[] = $this->io->collectString();
        }
        $ret['sceneOffsets'] = $sceneOffsets;
        $ret['sceneNames'] = $sceneNames;

        $frameNumbers = array();
        $frameLabels = array();
        $frameLabelCount = $this->io->collectEncodedU32();
        for ($i = 0; $i < $frameLabelCount; $i++) {
            $frameNumbers[] = $this->io->collectEncodedU32();
            $frameLabels[] = $this->io->collectString();
        }
        $ret['frameNumbers'] = $frameNumbers;
        $ret['frameLabels'] = $frameLabels;

        return $ret;
    }

    private function parseDefineBinaryDataTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineBinaryData';
        $ret['tag'] = $this->io->collectUI16();
        $this->io->collectUI32(); // Reserved, must be 0
        $ret['data'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        return $ret;
    }

    private function parseDefineFontNameTag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineFontName';
        $ret['fontId'] = $this->io->collectUI16();
        $ret['fontName'] = $this->io->collectString();
        $ret['fontCopyright'] = $this->io->collectString();
        return $ret;
    }

    private function parseDefineFont4Tag($bytePosEnd) {
        $ret = array();
        $ret['tagType'] = 'DefineFont4';
        $ret['fontId'] = $this->io->collectUI16();

        $this->io->collectUB(5); // Reserved
        $ret['fontFlagsHasFontData'] = $this->io->collectUB(1);
        $ret['fontFlagsItalic'] = $this->io->collectUB(1);
        $ret['fontFlagsBold'] = $this->io->collectUB(1);

        $ret['fontName'] = $this->io->collectString();

        if ($ret['fontFlagsHasFontData'] == 1) {
            $ret['fontData'] = $this->io->collectBytes($bytePosEnd - $this->io->bytePos);
        }
        return $ret;
    }
}