<?php

namespace swf2ass\ass;

use swf2ass\Constants;
use swf2ass\math\RealNumber;
use swf2ass\MatrixTransform;
use swf2ass\Vector2;

class matrixTransformTag implements ASSPositioningTag {

    private scaleTag $scale;
    private rotationTag $rotation;
    private shearingTag $shear;
    private MatrixTransform $transform;

    public function __construct(MatrixTransform $transform, Vector2 $scale, $rotationX, $rotationY, $rotationZ, $shearX, $shearY) {
        $this->transform = $transform;
        $this->scale = new scaleTag($scale);
        $this->rotation = new rotationTag($rotationX, $rotationY, $rotationZ);
        $this->shear = new shearingTag(new Vector2($shearX, $shearY));
    }

    public function transitionMatrixTransform(ASSLine $line, MatrixTransform $transform): ?matrixTransformTag {
        return self::fromMatrixTransform($transform);
    }

    public function encode(ASSEventTime $event): string {
        return /*sprintf("\\matrix(%.05F,%.05F,%.05F,%.05F,%.02F,%.02F)", $this->transform->get_a()->toFloat(), $this->transform->get_b()->toFloat(), $this->transform->get_c()->toFloat(), $this->transform->get_d()->toFloat(), $this->transform->get_tx()->divide(Constants::TWIP_SIZE)->toFloat(), $this->transform->get_ty()->divide(Constants::TWIP_SIZE)->toFloat())
            . */$this->scale->encode($event) . $this->rotation->encode($event) . $this->shear->encode($event);
    }

    public function equals(ASSTag $tag): bool {
        return $tag instanceof $this and $this->transform->equals($tag->transform) and $this->scale->equals($tag->scale) and $this->rotation->equals($tag->rotation) and $this->shear->equals($tag->shear);
    }

    private static function getByTransform_NumericallyStable(MatrixTransform $transform) : ?matrixTransformTag{

        //Numerically stable implementation by MrSmile

        $a = $transform->get_a();
        $b = $transform->get_b();
        $c = $transform->get_c();
        $d = $transform->get_d();

        $ac2 = $a->power(2)->add($c->power(2));
        $bd2 = $b->power(2)->add($d->power(2));

        $det = $a->multiply($d)->subtract($b->multiply($c));
        $dot = $a->multiply($b)->add($c->multiply($d));

        $scale_x = $scale_y = $frx = $fry = $frz = $fax = $fay = new RealNumber(0);

        if($ac2->greater($bd2)){
            if($ac2->greater(0)){
                $frz = new RealNumber(atan2($c->toFloat(), $a->toFloat()) * (180 / M_PI));
                $scale_x = $ac2->sqrt();
                $scale_y = $det->absolute()->divide($ac2->sqrt());
                $fax = $dot->divide($ac2);

                if($det->lesser(0)){
                    $frz = $frz->negate();
                    $frx = new RealNumber(180);
                }
            }
        }else{
            if($bd2->greater(0)){
                $frz = new RealNumber(atan2($b->negate()->toFloat(), $d->toFloat()) * (180 / M_PI));
                $scale_x = $det->absolute()->divide($bd2->sqrt());
                $scale_y = $bd2->sqrt();
                $fay = $dot->divide($bd2);

                if($det->lesser(0)){
                    $frz = $frz->negate();
                    $fry = new RealNumber(180);
                }
            }
        }

        $frz = $frz->negate();
        $fscx = $scale_x->absolute()->multiply(100);
        $fscy = $scale_y->absolute()->multiply(100);

        return new matrixTransformTag($transform, new Vector2($fscx->toFloat(), $fscy->toFloat()), $frx->toFloat(), $fry->toFloat(), $frz->toFloat(), $fax->toFloat(), $fay->toFloat());
    }


    private static function getByTransform_NumericallyUnstable(MatrixTransform $transform) : ?matrixTransformTag{

        //Numerically unstable implementation by Oneric

        $isZero = function (RealNumber $v) : bool{
            return $v->isZero(new RealNumber("0.000001")); ///TODO
        };

        $a = $transform->get_a();
        $b = $transform->get_b();
        $c = $transform->get_c();
        $d = $transform->get_d();

        $scale_x = $scale_y = $frx = $fry = $frz = $fax = $fay = new RealNumber(0);

        if(
            !(
                ($isZero($a) and !$isZero($b))
                or
                ($isZero($d) and !$isZero($c))
            )
        ){
            //Trivial case
            $scale_x = $a;
            $scale_y = $d;

            $fax = $isZero($a) ? new RealNumber(0) : $b->divide($a);
            $fay = $isZero($d) ? new RealNumber(0) : $c->divide($d);

            if($scale_x->lesser(0)){
                $fry = new RealNumber(180);
            }
            if($scale_y->lesser(0)){
                $frx = new RealNumber(180);
            }
        }elseif (
            !(
                ($isZero($b) and !$isZero($a))
                or
                ($isZero($c) and !$isZero($d))
            )
        ){
            //Rowswap

            $frz = new RealNumber(90);
            $scale_x = $c;
            $scale_y = $b->negate();

            $fax = $isZero($c) ? new RealNumber(0) : $d->divide($c);
            $fay = $isZero($b) ? new RealNumber(0) : $a->divide($b);

            if($scale_x->lesser(0)){
                $frx = new RealNumber(180);
            }
            if($scale_y->lesser(0)){
                $fry = new RealNumber(180);
            }
        }elseif (
            $isZero($a) and $isZero($c) and !$isZero($b) and !$isZero($d)
        ){
            //Zero Col Left
            $scale_y = new RealNumber(0);
            $fax = new RealNumber(0);
            $fay = new RealNumber(0);
            $scale_x = $b->power(2)->add($d->power(2))->sqrt();
            $frz = new RealNumber(atan($b->negate()->toFloat() / $d->toFloat()) * (180 / M_PI));
            if($a->lesser(0)){ // atan always yields positive cos
                $frz = $frz->add(180);
            }
        }elseif (
            !$isZero($a) and !$isZero($c) and $isZero($b) and $isZero($d)
        ){
            //Zero Col Right
            $scale_y = new RealNumber(0);
            $fax = new RealNumber(0);
            $fay = new RealNumber(0);
            $scale_x = $a->power(2)->add($c->power(2))->sqrt();
            $frz = new RealNumber(atan($c->toFloat() / $a->toFloat()) * (180 / M_PI));
            if($a->lesser(0)){ // atan always yields positive cos
                $frz->frz = $frz->add(180);
            }
        }else {
            echo $transform . "\n";
            throw new \Exception("Invalid transform state. This should not happen");
        }

        $frz = $frz->negate();

        $fscx = $scale_x->absolute()->multiply(100);
        $fscy = $scale_y->absolute()->multiply(100);

        return new matrixTransformTag($transform, new Vector2($fscx->toFloat(), $fscy->toFloat()), $frx->toFloat(), $fry->toFloat(), $frz->toFloat(), $fax->toFloat(), $fay->toFloat());
    }

    public static function fromMatrixTransform(MatrixTransform $transform): ?matrixTransformTag {
        return self::getByTransform_NumericallyStable($transform);
    }
}