<?php

namespace swf2ass\math;

use Decimal\Decimal;
use MathPHP\Number\ObjectArithmetic;

class RealNumber implements ObjectArithmetic {

    private static ?RealNumber $pi = null;
    private static ?RealNumber $epsilon = null;
    private Decimal $decimal;
    const DECIMAL_PRECISION = Decimal::DEFAULT_PRECISION;
    /**
     * @param RealNumber|Decimal|int|float|string $value
     */
    public function __construct($value){
        if($value instanceof Decimal){
            $this->decimal = $value;
        }else if($value instanceof RealNumber){
            $this->decimal = $value->decimal;
        }else if(is_float($value)){
            $this->decimal = new Decimal(sprintf("%.0" . PHP_FLOAT_DIG . "F", $value));
        }else if(is_string($value) or is_integer($value)){
            $this->decimal = new Decimal($value, self::DECIMAL_PRECISION + 1);
        }else{
            $this->decimal = new Decimal($value, self::DECIMAL_PRECISION + 1);
        }
    }

    public static function pi() : RealNumber{
        return self::$pi ?? (self::$pi = new RealNumber("3.1415926535897932384626433832"));
    }

    public static function setDefaultEpsilon(RealNumber $number){
        self::$epsilon = $number;
    }

    public static function epsilon() :RealNumber{
        return self::$epsilon ?? (self::$epsilon = new RealNumber("0.0000000000001"));
    }

    public function greater($object_or_scalar) : bool{
        return $this->compareTo($object_or_scalar) === 1;
    }

    public function equalWithEpsilon($object_or_scalar, ?RealNumber $epsilon = null) : bool{
        if($epsilon === null){
            $epsilon = self::epsilon();
        }
        return $this->subtract($object_or_scalar)->absolute()->lesserEqual($epsilon);
    }

    public function isZero(?RealNumber $epsilon = null) : bool{
        return $this->equalWithEpsilon(0, $epsilon);//$this->decimal->isZero();
    }

    public function equal($object_or_scalar) : bool{
        return $this->compareTo($object_or_scalar) === 0;
    }

    public function lesser($object_or_scalar) : bool{
        return $this->compareTo($object_or_scalar) === -1;
    }

    public function greaterEqual($object_or_scalar) : bool{
        return $this->compareTo($object_or_scalar) >= 0;
    }

    public function lesserEqual($object_or_scalar) : bool{
        return $this->compareTo($object_or_scalar) <= 0;
    }

    public function compareTo($object_or_scalar) : int{
        if($object_or_scalar instanceof RealNumber){
            return $this->decimal->compareTo($object_or_scalar->decimal);
        }else if($object_or_scalar instanceof Decimal){
            return $this->decimal->compareTo($object_or_scalar);
        }else{
            return $this->decimal->compareTo((new RealNumber($object_or_scalar))->decimal);
        }
    }

    public function negate() : RealNumber{
        return new RealNumber($this->decimal->negate());
    }

    public function cosine(int $terms = 12) : RealNumber{
        if($this->decimal->isZero()){
            return new RealNumber(1);
        }

        $div = $this->divide(self::pi())->toInteger();
        $x = $this->subtract(self::pi()->multiply($div));
        $sign = 1;
        if($div % 2 !== 0){
            $sign = -1;
        }

        $result = new RealNumber(1);
        $inter = new RealNumber(1);
        $num = $x->power(2);
        for($i = 1; $i <= $terms; ++$i){
            $comp = 2 * $i;
            $den = $comp * ($comp - 1);
            $inter = $inter->multiply($num->divide($den));
            if($i % 2 === 0){
                $result = $result->add($inter);
            }else{
                $result = $result->subtract($inter);
            }
        }

        return $result->multiply($sign);
    }

    public function sine(int $terms = 12) : RealNumber{
        if($this->decimal->isZero()){
            return new RealNumber(0);
        }
        return self::pi()->divide(2)->subtract($this)->cosine($terms);
    }

    public function getDecimal() : Decimal{
        return $this->decimal;
    }

    public function toFloat() : float{
        return $this->decimal->toFloat();
    }

    public function toFixed($places = self::DECIMAL_PRECISION) : string{
        return $this->decimal->trim()->toFixed($places);
    }

    public function toInteger() : int{
        return $this->decimal->toInt();
    }

    public function __toString(){
        return $this->toFixed();
    }

    public function __debugInfo(){
        return [
            "precision" => $this->getDecimal()->precision(),
            "value" => $this->toFixed()
        ];
    }


    public function add($object_or_scalar): RealNumber {
        if($object_or_scalar instanceof RealNumber){
            return new RealNumber($this->decimal->add($object_or_scalar->decimal));
        }else if($object_or_scalar instanceof Decimal){
            return new RealNumber($this->decimal->add($object_or_scalar));
        }else{
            return new RealNumber($this->decimal->add((new RealNumber($object_or_scalar))->decimal));
        }
    }

    public function subtract($object_or_scalar): RealNumber {
        if($object_or_scalar instanceof RealNumber){
            return new RealNumber($this->decimal->sub($object_or_scalar->decimal));
        }else if($object_or_scalar instanceof Decimal){
            return new RealNumber($this->decimal->sub($object_or_scalar));
        }else{
            return new RealNumber($this->decimal->sub((new RealNumber($object_or_scalar))->decimal));
        }
    }

    public function absolute(): RealNumber {
        return new RealNumber($this->decimal->abs());
    }

    public function sqrt(): RealNumber {
        return new RealNumber($this->decimal->sqrt());
    }

    public function power($object_or_scalar): RealNumber {
        if($object_or_scalar instanceof RealNumber){
            return new RealNumber($this->decimal->pow($object_or_scalar->decimal));
        }else if($object_or_scalar instanceof Decimal){
            return new RealNumber($this->decimal->pow($object_or_scalar));
        }else{
            return new RealNumber($this->decimal->pow((new RealNumber($object_or_scalar))->decimal));
        }
    }

    public function multiply($object_or_scalar): RealNumber {
        if($object_or_scalar instanceof RealNumber){
            return new RealNumber($this->decimal->mul($object_or_scalar->decimal));
        }else if($object_or_scalar instanceof Decimal){
            return new RealNumber($this->decimal->mul($object_or_scalar));
        }else{
            return new RealNumber($this->decimal->mul((new RealNumber($object_or_scalar))->decimal));
        }
    }

    public function divide($object_or_scalar): RealNumber {
        if($object_or_scalar instanceof RealNumber){
            return new RealNumber($this->decimal->div($object_or_scalar->decimal));
        }else if($object_or_scalar instanceof Decimal){
            return new RealNumber($this->decimal->div($object_or_scalar));
        }else{
            return new RealNumber($this->decimal->div((new RealNumber($object_or_scalar))->decimal));
        }
    }

    public static function createZeroValue(): RealNumber {
        return new RealNumber(0);
    }
}